#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import, division, with_statement

import copy
import logging
import os
import time

import pymysql

logger = logging.getLogger(__name__)

class Connection(object):
    """A lightweight wrapper around PyMySQL DB-API connections.

    The main value we provide is wrapping rows in a dict/object so that
    columns can be accessed by name.

    Cursors are hidden by the implementation, but other than that, the methods
    are very similar to the DB-API.

    We explicitly set the timezone to UTC and assume the character encoding to
    UTF-8 (can be changed) on all connections to avoid time zone and encoding errors.

    The sql_mode parameter is set by default to "traditional", which "gives an error instead of a warning"
    (http://dev.mysql.com/doc/refman/5.0/en/server-sql-mode.html). However, it can be set to
    any other mode including blank (None) thereby explicitly clearing the SQL mode.

    Arguments read_timeout and write_timeout can be passed using kwargs, if MySQL version > 5.1.12.
    """
    def __init__(self, host, database, user=None, password=None,
                 max_idle_time=7 * 3600, connect_timeout=0,
                 time_zone="+0:00", charset = "utf8", sql_mode="TRADITIONAL",
                 **kwargs):
        self.host = host
        self.database = database
        self.max_idle_time = float(max_idle_time)

        args = dict(conv=CONVERSIONS, use_unicode=True, charset=charset,
                    db=database, init_command=('SET time_zone = "%s"' % time_zone),
                    connect_timeout=connect_timeout, sql_mode=sql_mode, **kwargs)
        if user is not None:
            args["user"] = user
        if password is not None:
            args["passwd"] = password

        # We accept a path to a MySQL socket file or a host(:port) string
        if "/" in host:
            args["unix_socket"] = host
        else:
            self.socket = None
            pair = host.split(":")
            if len(pair) == 2:
                args["host"] = pair[0]
                args["port"] = int(pair[1])
            else:
                args["host"] = host
                args["port"] = 3306

        self._db = None
        self._db_args = args
        self._last_use_time = time.time()
        try:
            self.reconnect()
        except Exception:
            logger.error("Cannot connect to MySQL on %s", self.host,
                          exc_info=True)

    def __del__(self):
        self.close()

    def close(self):
        """Closes this database connection."""
        if getattr(self, "_db", None) is not None:
            self._db.close()
            self._db = None

    def reconnect(self):
        """Closes the existing database connection and re-opens it."""
        self.close()
        self._db = pymysql.connect(**self._db_args)
        self._db.autocommit(True)

    def iter(self, query, *parameters, **kwparameters):
        """Returns an iterator for the given query and parameters."""
        self._ensure_connected()
        cursor = pymysql.cursors.SSCursor(self._db)
        try:
            self._execute(cursor, query, parameters, kwparameters)
            column_names = [d[0] for d in cursor.description]
            for row in cursor:
                yield Row(zip(column_names, row))
        finally:
            cursor.close()

    def query(self, query, *parameters, **kwparameters):
        """Returns a row list for the given query and parameters."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            column_names = [d[0] for d in cursor.description]
            return [Row(zip(column_names, row)) for row in cursor]
        finally:
            cursor.close()

    def get(self, query, *parameters, **kwparameters):
        """Returns the (singular) row returned by the given query.

        If the query has no results, returns None.  If it has
        more than one result, raises an exception.
        """
        rows = self.query(query, *parameters, **kwparameters)
        if not rows:
            return None
        elif len(rows) > 1:
            raise Exception("Multiple rows returned for Database.get() query")
        else:
            return rows[0]

    # rowcount is a more reasonable default return value than lastrowid,
    # but for historical compatibility execute() must return lastrowid.
    def execute(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        return self.execute_lastrowid(query, *parameters, **kwparameters)

    def execute_lastrowid(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            return cursor.lastrowid
        finally:
            cursor.close()

    def execute_rowcount(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the rowcount from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            return cursor.rowcount
        finally:
            cursor.close()

    def executemany(self, query, parameters):
        """Executes the given query against all the given param sequences.

        We return the lastrowid from the query.
        """
        return self.executemany_lastrowid(query, parameters)

    def executemany_lastrowid(self, query, parameters):
        """Executes the given query against all the given param sequences.

        We return the lastrowid from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            # return cursor.lastrowid
        finally:
            cursor.close()

    def executemany_rowcount(self, query, parameters):
        """Executes the given query against all the given param sequences.

        We return the rowcount from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            return cursor.rowcount
        finally:
            cursor.close()

    update = delete = execute_rowcount
    updatemany = executemany_rowcount

    insert = execute_lastrowid
    insertmany = executemany_lastrowid

    def _ensure_connected(self):
        # Mysql by default closes client connections that are idle for
        # 8 hours, but the client library does not report this fact until
        # you try to perform a query and it fails.  Protect against this
        # case by preemptively closing and reopening the connection
        # if it has been idle for too long (7 hours by default).
        if (self._db is None or
            (time.time() - self._last_use_time > self.max_idle_time)):
            self.reconnect()
        self._last_use_time = time.time()

    def _cursor(self):
        self._ensure_connected()
        return self._db.cursor()

    def _execute(self, cursor, query, parameters, kwparameters):
        try:
            return cursor.execute(query, kwparameters or parameters)
        except OperationalError:
            logger.error("Error connecting to MySQL on %s", self.host)
            self.close()
            raise


class Row(dict):
    """A dict that allows for object-like property access syntax."""
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

CONVERSIONS = copy.copy(pymysql.converters.conversions)

# Alias some common MySQL exceptions
IntegrityError = pymysql.IntegrityError
OperationalError = pymysql.OperationalError



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import logging
import smtplib
import time
from datetime import datetime, timedelta
from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.utils import COMMASPACE
from email.utils import formatdate

from tornado.escape import utf8
from tornado.options import options

__all__ = ("send_email", "EmailAddress")

# borrow email re pattern from django
_email_re = re.compile(
    # dot-atom
    r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
    # quoted-string
    r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"'
    r')@(?:[A-Z0-9]+(?:-*[A-Z0-9]+)*\.)+[A-Z]{2,6}$', re.IGNORECASE)  # domain


logger = logging.getLogger(__name__)


def send_email(fr, to, subject, body, html=None, attachments=[]):
    """Send an email.

    If an HTML string is given, a mulitpart message will be generated with
    plain text and HTML parts. Attachments can be added by providing as a
    list of (filename, data) tuples.
    """
    # convert EmailAddress to pure string
    if isinstance(fr, EmailAddress):
        fr = str(fr)
    else:
        fr = utf8(fr)
    to = [utf8(t) for t in to]

    if html:
        # Multipart HTML and plain text
        message = MIMEMultipart("alternative")
        message.attach(MIMEText(body, "plain"))
        message.attach(MIMEText(html, "html"))
    else:
        # Plain text
        message = MIMEText(body)
    if attachments:
        part = message
        message = MIMEMultipart("mixed")
        message.attach(part)
        for filename, data in attachments:
            part = MIMEBase("application", "octet-stream")
            part.set_payload(data)
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", "attachment",
                            filename=filename)
            message.attach(part)

    message["Date"] = formatdate(time.time())
    message["From"] = fr
    message["To"] = COMMASPACE.join(to)
    message["Subject"] = utf8(subject)

    _get_session().send_mail(fr, to, utf8(message.as_string()))


class EmailAddress(object):
    def __init__(self, addr, name=""):
        assert _email_re.match(addr), "Email address(%s) is invalid." % addr

        self.addr = addr
        if name:
            self.name = name
        else:
            self.name = addr.split("@")[0]

    def __str__(self):
        return '%s <%s>' % (utf8(self.name), utf8(self.addr))


class _SMTPSession(object):
    def __init__(self, host, user='', password='', duration=30, tls=False):
        self.host = host
        self.user = user
        self.password = password
        self.duration = duration
        self.tls = tls
        self.session = None
        self.deadline = datetime.now()
        self.renew()

    def send_mail(self, fr, to, message):
        if self.timeout:
            self.renew()

        try:
            self.session.sendmail(fr, to, message)
        except Exception as e:
            err = "Send email from %s to %s failed!\n Exception: %s!"
            logger.error(err, fr, to, e)
            self.renew()

    @property
    def timeout(self):
        if datetime.now() < self.deadline:
            return False
        else:
            return True

    def renew(self):
        try:
            if self.session:
                self.session.quit()
        except Exception:
            pass

        self.session = smtplib.SMTP(self.host)
        if self.user and self.password:
            if self.tls:
                self.session.starttls()

            self.session.login(self.user, self.password)

        self.deadline = datetime.now() + timedelta(seconds=self.duration * 60)


def _get_session():
    global _session
    if _session is None:
        _session = _SMTPSession(options.smtp['host'],
                                options.smtp['user'],
                                options.smtp['password'],
                                options.smtp['duration'],
                                options.smtp['tls'])

    return _session


_session = None



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility library"""

import re
import random

from six.moves import range
from tornado.web import url
from tornado.options import options
from tornado import escape

from djinn.compat import import_module
from djinn.db import Row
from djinn.errors import TemplateContextError


ALPHABET_FULL = '23456789abcdefghijkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ'
ALPHABET_LOWER_ONLY = '23456789abcdefghijkmnpqrstuvwxyz'


class Context(dict):

    """Template context container.

    A container which will return empty string silently if the key is not exist
    rather than raise AttributeError when get a item's value.

    It will raise TemplateContextError if debug is True and the key
    does not exist.

    The context item also can be accessed through get attribute.

    """

    def __setattr__(self, key, value):
        self[key] = value

    def __str__(self):
        return str(self)

    def __iter__(self):
        return iter(self.items())

    def __getattr__(self, key):
        """Get a context attribute.

        Raise TemplateContextError if the attribute not be set to
        avoid confused AttributeError exception when debug is True, or return ""

        """
        if key in self:
            return self[key]
        elif options.debug:
            raise TemplateContextError("'%s' does not exist in context" % key)
        else:
            return ""

    def __hasattr__(self, key):
        if key in self:
            return True
        else:
            return False


def load_url_handlers(handlers_root_module, handlers_modules,
                      prefix=""):
    """Load URL handlers

    ``handlers_root_module``: The root module of all of handlers
    ``handlers_modules``: A list of module which is the container of a
    suite of handlers
    ``prefix``: URL prefix
    """
    handlers = []
    for handlers_module in handlers_modules:
        module = import_module(".%s" % handlers_module, handlers_root_module)
        module_hanlders = getattr(module, "handlers", None)
        if module_hanlders:
            _handlers = []
            for handler in module_hanlders:
                try:
                    patten = r"%s%s" % (prefix, handler[0])
                    if len(handler) == 2:
                        _handlers.append((patten,
                                          handler[1]))
                    elif len(handler) == 3:
                        _handlers.append(url(patten,
                                             handler[1],
                                             name=handler[2])
                                         )
                    else:
                        pass
                except IndexError:
                    pass

            handlers.extend(_handlers)

    return handlers


def get_count(row):
    """Get count from a query result

    ``row``: Row object of a db query result
    """
    if row:
        count = row.c  # c as count by name convention
    else:
        count = 0

    return count


def get_column_values(rows, column=None):
    """Get all values of a column from a query result

    ``rows``: A list of Row object of a db query result
    ``column``: Column name which value will be collected
    """
    result = []
    if rows:
        for row in rows:
            if column:
                result.append(row.get(column, None))
            else:
                result.append(row.values()[0])

    return result


def pick_row_attrs(row, attrs):
    if not row or not isinstance(row, dict):
        return row

    row_ = Row()
    for name in attrs:
        row_[name] = row.get(name, None)

    return row_


def gen_uid(size=6, alphabet=ALPHABET_LOWER_ONLY):
    s = []
    # random part of uid
    for i in range(0, size):
        s.append(random.choice(alphabet))

    return "".join(s)


def parse_on_off(status, default="on"):
    if not status:
        return default

    return "on" if status == "on" else "off"


def shorten_content(content, length):
    content = content.strip(" \n\t")
    content = content if len(
        content) <= length else "%s..." % content[0:length - 3]
    return content


def compare_version(v_a, v_b):
    """compare version numbers

    normal cases:
    1.10.2 > 1.1.3
    1.2.1 < 1.2.2
    1.2.0000 = 1.2.0

    exception cases:
    1.2.1abc > 1.2.1
    None < 1.0
    "  " < 1.0
    """
    def normalize(v):
        vv = []
        v = escape.utf8(v)

        if v:
            for x in re.sub(r'(\.0+)*$', '', v).split("."):
                try:
                    x = int(x)  # 001 to 1
                    vv.append(x)
                except ValueError:
                    # may have characters in the tail sometimes
                    vv.append(x)

        return vv

    return cmp(normalize(v_a), normalize(v_b))



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import traceback
import logging

try:
    import ujson as json
except ImportError:
    import json

from six import text_type
from tornado import escape
from tornado.options import options
from tornado.web import RequestHandler as BaseRequestHandler, HTTPError
from djinn import errors
from djinn.utils import Context

REMOVE_SLASH_RE = re.compile(".+/$")

logger = logging.getLogger(__name__)


class BaseHandler(BaseRequestHandler):
    def prepare(self):
        self.remove_slash()
        self.prepare_context()
        self.traffic_threshold()

    def traffic_threshold(self):
        pass

    def prepare_context(self):
        self._context = Context()

    def remove_slash(self):
        if self.request.method == "GET":
            if REMOVE_SLASH_RE.match(self.request.path):
                # remove trail slash in path
                uri = self.request.path.rstrip("/")
                if self.request.query:
                    uri += "?" + self.request.query

                self.redirect(uri)

    def render_string(self, template_name, **kwargs):
        """Override default render_string, add context to template."""
        assert "context" not in kwargs, "context is a reserved word for \
                template context valuable."
        kwargs['context'] = self._context
        kwargs['url_escape'] = escape.url_escape

        return super(BaseHandler, self).render_string(template_name, **kwargs)

    def get_argument(self, name, default=BaseRequestHandler._ARG_DEFAULT, strip=True, as_bytes=True):
        value = super(BaseHandler, self).get_argument(name, default, strip)
        return escape.utf8(value) if as_bytes and isinstance(value, text_type) else value

    def get_int_argument(self, name, default=0):
        try:
            return int(self.get_argument(name, default))
        except ValueError:
            return default


class APIHandler(BaseHandler):
    def finish(self, chunk=None, message=None):
        _ = self.locale.translate
        if chunk is None:
            chunk = {}

        if isinstance(chunk, dict):
            chunk = {"meta": {"code": errors.HTTP_OK}, "data": chunk}

            if message:
                chunk["message"] = _(message)
        elif isinstance(chunk, errors.HTTPAPIError):
            chunk.message = _(chunk.message)
            chunk = str(chunk)

        callback = escape.utf8(self.get_argument("callback", None))
        if callback:
            self.set_header("Content-Type", "application/x-javascript")

            if isinstance(chunk, dict):
                chunk = escape.json_encode(chunk)

            self._write_buffer = [callback, "(", chunk, ")"] if chunk else []
            super(APIHandler, self).finish()
        else:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            if isinstance(chunk, dict):
                chunk = json.dumps(chunk, ensure_ascii=False, sort_keys=True)
            super(APIHandler, self).finish(chunk)

    def write_error(self, status_code, **kwargs):
        try:
            exc_info = kwargs.pop('exc_info')
            e = exc_info[1]

            if isinstance(e, errors.HTTPAPIError):
                pass
            elif isinstance(e, HTTPError):
                e = errors.HTTPAPIError(e.status_code, message=e.log_message)
            else:
                e = errors.HTTPAPIError(errors.ERROR_INTERNAL_SERVER_ERROR)

            exception = "".join([ln for ln
                                 in traceback.format_exception(*exc_info)])

            if status_code == errors.ERROR_INTERNAL_SERVER_ERROR \
                    and not options.debug:
                self.send_error_mail(exception)

            self.clear()
            # always return 200 OK for API errors
            self.set_status(errors.HTTP_OK)
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.finish(e)
        except Exception:
            logger.error(traceback.format_exc())
            return super(APIHandler, self).write_error(status_code, **kwargs)

    def send_error_mail(self, exception):
        """Override to implement custom error mail"""
        pass


class ErrorHandler(BaseHandler):
    """Default 404: Not Found handler."""

    def prepare(self):
        super(ErrorHandler, self).prepare()
        raise HTTPError(errors.ERROR_NOT_FOUND)


class APIErrorHandler(APIHandler):
    """Default API 404: Not Found handler."""

    def prepare(self):
        super(APIErrorHandler, self).prepare()
        raise errors.HTTPAPIError(errors.ERROR_NOT_FOUND)



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base Application"""

import os

from tornado.locale import load_translations
from tornado.web import Application


class DjinnApplication(Application):
    """Base Application for djinn framework"""

    def reverse_api(self, request):
        """Returns a URL name for a request"""
        handlers = self._get_host_handlers(request)

        for spec in handlers:
            match = spec.regex.match(request.path)
            if match:
                return spec.name

        return None


translation_folder = "translations"
here = os.path.dirname(os.path.realpath(__file__))
load_translations(os.path.join(here, translation_folder))

app_translation = os.path.join(os.getcwd(), translation_folder)
if os.path.isdir(app_translation):
    load_translations(app_translation)



# -*- coding: utf-8 -*-

from importlib import import_module as _import

from six import PY3

def import_module(name, package=None):
	if PY3 and package is not None:
		_import(package)
	return _import(name, package=package)



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""tornado.options enhancement"""

import io
import traceback
import os

from tornado.options import parse_command_line, options, define


def parse_config_file(filename):
    """Rewrite tornado default parse_config_file.

    Parses and loads the Python config file at the given filename.
    This version allow customize new options which are not defined before
    from a configuration file.
    """
    config = {}
    exec(compile(io.open(filename, encoding="UTF-8").read().encode("UTF-8"), filename, "exec"), {}, config)

    for name in config:
        if name in options._options:
            options._options[name].set(config[name])
        else:
            define(name, config[name])


def parse_options(root_dir, settings_file="settings", parse_cmd=True):
    """Parse options file and command line"""
    try:
        parse_config_file(os.path.join(root_dir, "%s.py" % settings_file))
        # print "Using settings.py as default settings."
    except Exception as exc:
        print("No any default settings, are you sure? Exception: %s" % exc)

    try:
        parse_config_file(
            os.path.join(root_dir, "%s_local.py" % settings_file))
        # print "Override some settings with local settings."
    except Exception as exc:
        print("No local settings. Exception: %s" % exc)
        # print traceback.format_exc()

    if parse_cmd:
        parse_command_line()



# -*- coding: utf-8 -*-
#
# Copyright(c) 2014 palmhold.com
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import logging
import random

from djinn.db import Connection as BaseConnection
from six import iteritems, PY2
from tornado.options import options, define
from tornado import escape

from ..errors import DatastoreError

define("log_db_query", True, bool, "whether log db query statements")

manager = None

logger = logging.getLogger(__name__)

def setup(datastore_pool):
    global manager

    if manager is None:
        manager = MysqlManager(datastore_pool)
    return manager


class MysqlManager(object):
    _datastore_pool = {}

    def __init__(self, datastore_pool):
        for k, v in iteritems(datastore_pool):
            MysqlManager._datastore_pool[k] = MysqlMSConnection(v[0], v[1])

    def __getattr__(self, instance):
        r = self._datastore_pool.get(instance, None)
        if not r:
            raise DatastoreError("Mysql %s instance does not exist"
                                 % instance)

        return r


class MysqlMSConnection(object):

    """Mysql master & slave connection

    Manage datastore connection instances, it selects master or slave instance
    according to SQL statement automatically.

    """

    def __init__(self, master, slaves):
        self.master = Connection(master)

        self._slave_conns = []
        for slave in slaves:
            self._slave_conns.append(Connection(slave))

    @property
    def query(self):
        return self.slave.query

    @property
    def get(self):
        return self.slave.get

    @property
    def execute(self):
        return self.master.execute

    @property
    def executemany(self):
        return self.master.executemany

    @property
    def insert(self):
        return self.master.insert

    @property
    def update(self):
        return self.master.update

    @property
    def slave(self):
        return random.choice(self._slave_conns)


class Connection(BaseConnection):

    def __init__(self, options):
        default_options = {
            "host": "localhost:3306",
            "database": "test",
            "user": None,
            "password": None,
            "charset": "utf8mb4",
            "max_idle_time": 7 * 3600,
            "connect_timeout": 10,
            "time_zone": "+8:00"
        }
        default_options.update(options)

        super(Connection, self).__init__(**default_options)

    def _execute(self, cursor, query, parameters, kwparameters):
        # Override default _execute to log executing info when log_query is on
        parameters_ = []
        for parameter in parameters:
            if isinstance(parameter, unicode if PY2 else str):
                parameter = escape.utf8(parameter)
            parameters_.append(parameter)
        parameters = tuple(parameters_)

        if options.log_db_query:
            sql = query % parameters
            try:
                start = time.time()
                r = super(Connection, self)._execute(cursor, query, parameters,
                                                     kwparameters)
                elapse = time.time() - start
                logger.debug("SQL executing elapse %s seconds on %s: %s",
                             elapse, self.host, sql)

                return r
            except Exception as e:
                logger.error("SQL: %s", sql)
                raise e
        else:
            return super(Connection, self)._execute(cursor, query, parameters, kwparameters)
